/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2010-2015 Intel Corporation
 */

#include <stdint.h>

#include <inttypes.h>
#include <signal.h>
#include <stdbool.h>
#include <getopt.h>
#include <rte_eal.h>
#include <rte_ethdev.h>
#include <rte_memcpy.h>
#include <rte_byteorder.h>
#include <rte_lcore.h>
#include <rte_mbuf.h>
#include <rte_version.h>
#include <rte_string_fns.h>
#include<stdio.h>
#include<sys/ipc.h>
#include<sys/shm.h>
#include<sys/types.h>
#include<string.h>
#include<errno.h>
#include<unistd.h>
#include<string.h>
#include <fcntl.h>
#include <sys/mman.h>


#include <sched.h>
#include <sys/resource.h>

#include <numa.h> // se non metto questo il numa_alloc non funziona e l'accesso a quelle zone di memoria fanno segmentation
#include <numaif.h>	

#include <rte_mempool.h>
#include <math.h>
#include <stdlib.h>

#include "test-jitter-dpdk.h"

#define RX_RING_SIZE 1024
#define TX_RING_SIZE 1024

#define NUM_MBUFS 8191
//#define NUM_MBUFS 16382
#define MBUF_CACHE_SIZE 250

#define NUM_SAMPLES 10000000

#define SHM 1
#define STORAGE_ID_SHM "/SHM_TEST1"
#define STORAGE_ID_HUGE "/mnt/hugepages/shm"
#define STORAGE_SIZE 10000000

/*
 * Configurable number of RX/TX ring descriptors
 */
#define RTE_TEST_RX_DESC_DEFAULT 1024
#define RTE_TEST_TX_DESC_DEFAULT 1024

#define MAX_TX_QUEUE_PER_PORT RTE_MAX_LCORE
#define MAX_RX_QUEUE_PER_PORT 128

#define MAX_LCORE_PARAMS 1024
#define MAX_STAT	128

/* Static global variables used within this file. */
static uint16_t nb_rxd = RTE_TEST_RX_DESC_DEFAULT;
static uint16_t nb_txd = RTE_TEST_TX_DESC_DEFAULT;

static uint16_t stat_id = 0;

static int numa_on = 1; /**< NUMA is enabled by default. */

volatile bool force_quit;

struct lcore_conf lcore_conf[RTE_MAX_LCORE];

struct lcore_params {
	uint16_t port_id;
	uint8_t queue_id;
	uint8_t lcore_id;
} __rte_cache_aligned;

static struct lcore_params lcore_params_array[MAX_LCORE_PARAMS];
static struct lcore_params lcore_params_array_default[] = {
	{0, 0, 2},
	{0, 1, 2},
	{0, 2, 2},
	{1, 0, 2},
	{1, 1, 2},
	{1, 2, 2},
	{2, 0, 2},
	{3, 0, 3},
	{3, 1, 3},
};

static struct lcore_params * lcore_params = lcore_params_array_default;
static uint16_t nb_lcore_params = sizeof(lcore_params_array_default) /
				sizeof(lcore_params_array_default[0]);

#pragma pack(push, 1)
struct rt_fpga_packet {
   unsigned char eth_cpu_mac[6]    ; // destination MAC 6
   unsigned char eth_fpga_mac[6] ; // source MAC 12
   unsigned short eth_type         ; // protocol type 0x0800 => IP 14

   unsigned char ip_ver_len        ; // IP {version, header length in 32 bit words} 15
   unsigned char ip_tos            ; // IP type of service 16
   unsigned short ip_len           ; // IP total length 18
   unsigned short ip_id            ; // IP datagram fragment identification -> not used since datagrams are not fragmented 20
   unsigned short ip_flags_fr_ofs  ; // IP datagram flags (don't fragment and last fragment), fragment offset 0x0 22
   unsigned char ip_ttl            ; // IP time to live -> max -> 0xFF 24
   unsigned char ip_prot           ; // IP protocol -> 0x11 -> UDP
   unsigned short ip_chksum        ; // IP header checksum 26
   unsigned char ip_fpga_ip[4]   ; // IP source IP address 30
   unsigned char ip_cpu_ip[4]      ; // IP destination IP address 34

   unsigned short udp_fpga_port  ; // UDP source port
   unsigned short udp_cpu_port     ; // UDP destination port
   unsigned short udp_length       ; // UDP length in bytes (data+header)
   unsigned short udp_chksum       ; // UDP checksum -> 0x0 -> disabled 42

   uint32_t Va			; // amplitude A 46
   uint32_t Vb			; // amplitude B 50



   uint32_t Vc			; // amplitude C
   uint32_t Vd			; // amplitude D
   uint32_t Sum			; // amplitudes' sum
   int32_t Q			; // q offset
   int32_t X			; // x offset
   int32_t Y			; // y offset
   uint64_t seq		;  // packet sequence field
   unsigned short stat		; // hw status
} rt_fpga_packet;
#pragma pack(pop)

/* mask of enabled ports */
uint32_t enabled_port_mask;


static const struct rte_eth_conf port_conf_default = {
	.rxmode = {
#if RTE_VERSION < RTE_VERSION_NUM(22,7,0,0)
		.mq_mode = ETH_MQ_RX_RSS,
		.offloads = DEV_RX_OFFLOAD_CHECKSUM,
#else
		.mq_mode = RTE_ETH_MQ_RX_RSS,
		.offloads = RTE_ETH_RX_OFFLOAD_CHECKSUM,
#endif
		.split_hdr_size = 0,
	},
	.rx_adv_conf = {
		.rss_conf = {
			.rss_key = NULL,
#if RTE_VERSION < RTE_VERSION_NUM(22,7,0,0)
			.rss_hf = ETH_RSS_UDP,
			//.rss_hf = ETH_RSS_IP,
#else
			.rss_hf = RTE_ETH_RSS_UDP,
			//.rss_hf = RTE_ETH_RSS_IP,
#endif
		},
	},
	.txmode = {
#if RTE_VERSION < RTE_VERSION_NUM(22,7,0,0)
		.mq_mode = ETH_MQ_TX_NONE,
#else
		.mq_mode = RTE_ETH_MQ_TX_NONE,
#endif
	},
};

#if RTE_VERSION >= RTE_VERSION_NUM(22,7,0,0)
static uint32_t max_pkt_len;
#endif

static struct rte_mempool *pktmbuf_pool[RTE_MAX_ETHPORTS][NB_SOCKETS];

struct rt_fpga_packet bpm_packet[3];

struct timespec old_timestamp, new_timestamp;

struct lcore_stat	qstat[MAX_STAT];

//	long long int min_time = 1e10, max_time = 0, tmp_time, max_num_packet = 0, lost_packet = 0, 
//	spurious_packet = 0, rx_packets = 0,
//	old_timestamp_tsc = 0, new_timestamp_tsc = 0, 
//	min_loop_time = 0, max_loop_time = 0,
//	rte_eth_rx_burst_max_time = 0,max_packet_time = 0,max_free_time = 0;

unsigned short old_seq = 0, cur_seq = 0;

unsigned short lock_flag = 0;

long long int cxus = 3100;

volatile bool stop_flag;
int32_t warmup_cycles = 20;

void print_bpm_packet(struct rt_fpga_packet *mpacket);

static int
parse_portmask(const char *portmask)
{
	char *end = NULL;
	unsigned long pm;

	/* parse hexadecimal string */
	pm = strtoul(portmask, &end, 16);
	if ((portmask[0] == '\0') || (end == NULL) || (*end != '\0'))
		return -1;

	if (pm == 0)
		return -1;

	return pm;
}

static int
parse_config(const char *q_arg)
{
	char s[256];
	const char *p, *p0 = q_arg;
	char *end;
	enum fieldnames {
		FLD_PORT = 0,
		FLD_QUEUE,
		FLD_LCORE,
		_NUM_FLD
	};
	unsigned long int_fld[_NUM_FLD];
	char *str_fld[_NUM_FLD];
	int i;
	unsigned size;

	nb_lcore_params = 0;

	while ((p = strchr(p0,'(')) != NULL) {
		++p;
		if((p0 = strchr(p,')')) == NULL)
			return -1;

		size = p0 - p;
		if(size >= sizeof(s))
			return -1;

		snprintf(s, sizeof(s), "%.*s", size, p);
		if (rte_strsplit(s, sizeof(s), str_fld, _NUM_FLD, ',') != _NUM_FLD)
			return -1;
		for (i = 0; i < _NUM_FLD; i++){
			errno = 0;
			int_fld[i] = strtoul(str_fld[i], &end, 0);
			if (errno != 0 || end == str_fld[i] || int_fld[i] > 255)
				return -1;
		}
		if (nb_lcore_params >= MAX_LCORE_PARAMS) {
			printf("exceeded max number of lcore params: %hu\n",
				nb_lcore_params);
			return -1;
		}
		lcore_params_array[nb_lcore_params].port_id =
			(uint8_t)int_fld[FLD_PORT];
		lcore_params_array[nb_lcore_params].queue_id =
			(uint8_t)int_fld[FLD_QUEUE];
		lcore_params_array[nb_lcore_params].lcore_id =
			(uint8_t)int_fld[FLD_LCORE];
		++nb_lcore_params;
	}
	lcore_params = lcore_params_array;
	return 0;
}

#define MAX_JUMBO_PKT_LEN  9600
#define MEMPOOL_CACHE_SIZE 256

static const char short_options[] =
	"p:"  /* portmask */
	;
#define CMD_LINE_OPT_CONFIG "config"

enum {
	/* long options mapped to a short option */

	/* first long only option value must be >= 256, so that we won't
	 * conflict with short options */
	CMD_LINE_OPT_MIN_NUM = 256,
	CMD_LINE_OPT_CONFIG_NUM
};

static const struct option lgopts[] = {
	{CMD_LINE_OPT_CONFIG, 1, 0, CMD_LINE_OPT_CONFIG_NUM},
	{NULL, 0, 0, 0}
};

static int
check_lcore_params(void)
{
	uint8_t queue, lcore;
	uint16_t i;
	int socketid;

	for (i = 0; i < nb_lcore_params; ++i) {
		queue = lcore_params[i].queue_id;
		if (queue >= MAX_RX_QUEUE_PER_PORT) {
			printf("invalid queue number: %hhu\n", queue);
			return -1;
		}
		lcore = lcore_params[i].lcore_id;
		if (!rte_lcore_is_enabled(lcore)) {
			printf("error: lcore %hhu is not enabled in lcore mask\n", lcore);
			return -1;
		}
		if ((socketid = rte_lcore_to_socket_id(lcore) != 0) &&
			(numa_on == 0)) {
			printf("warning: lcore %hhu is on socket %d with numa off \n",
				lcore, socketid);
		}
	}
	return 0;
}

static int
check_port_config(void)
{
	uint16_t portid;
	uint16_t i;

	for (i = 0; i < nb_lcore_params; ++i) {
		portid = lcore_params[i].port_id;
		if ((enabled_port_mask & (1 << portid)) == 0) {
			printf("port %u is not enabled in port mask\n", portid);
			return -1;
		}
		if (!rte_eth_dev_is_valid_port(portid)) {
			printf("port %u is not present on the board\n", portid);
			return -1;
		}
	}
	return 0;
}

static uint8_t
get_port_n_rx_queues(const uint16_t port)
{
	int queue = -1;
	uint16_t i;

	for (i = 0; i < nb_lcore_params; ++i) {
		if (lcore_params[i].port_id == port) {
			if (lcore_params[i].queue_id == queue+1)
				queue = lcore_params[i].queue_id;
			else
				rte_exit(EXIT_FAILURE, "queue ids of the port %d must be"
						" in sequence and must start with 0\n",
						lcore_params[i].port_id);
		}
	}
	return (uint8_t)(++queue);
}

static int
init_lcore_rx_queues(void)
{
	uint16_t i, nb_rx_queue;
	uint8_t lcore;

	for (i = 0; i < nb_lcore_params; ++i) {
		lcore = lcore_params[i].lcore_id;
		nb_rx_queue = lcore_conf[lcore].n_rx_queue;
		if (nb_rx_queue >= MAX_RX_QUEUE_PER_LCORE) {
			printf("error: too many queues (%u) for lcore: %u\n",
				(unsigned)nb_rx_queue + 1, (unsigned)lcore);
			return -1;
		} else {
			lcore_conf[lcore].rx_queue_list[nb_rx_queue].port_id =
				lcore_params[i].port_id;
			lcore_conf[lcore].rx_queue_list[nb_rx_queue].queue_id =
				lcore_params[i].queue_id;
			lcore_conf[lcore].n_rx_queue++;
			lcore_conf[lcore].stat_id = stat_id++;
		}
	}
	return 0;
}

/* display usage */
static void
print_usage(const char *prgname)
{
	fprintf(stderr, "%s [EAL options] --"
		" -p PORTMASK"
		" --config (port,queue,lcore)[,(port,queue,lcore)]\n\n"
		"  -p PORTMASK: Hexadecimal bitmask of ports to configure\n"
		"  --config (port,queue,lcore): Rx queue configuration\n",
		prgname);
}


/*
 * Custom handling of signals to handle stats 
 */
static void
signal_handler(int signum)
{
	//double X=1;

	/* When we receive a USR1 signal, print stats */
	if (signum == SIGUSR1) {

		// load CPU
		//for(int i = 0; i < 10000; i++) {
		//	X = pow(sin((double)(rand() * 1000)), X);
		//}
		// end load CPU

		stop_flag = true;
	}
	else if (signum == SIGINT || signum == SIGTERM) {
		printf("\n\nSignal %d received, preparing to exit...\n",
				signum);
		force_quit = true;
	}

}

#if RTE_VERSION >= RTE_VERSION_NUM(22,7,0,0)
static int
config_port_max_pkt_len(struct rte_eth_conf *conf,
		struct rte_eth_dev_info *dev_info)
{
	uint32_t overhead_len;

	if (max_pkt_len == 0)
		return 0;

	if (max_pkt_len < RTE_ETHER_MIN_LEN || max_pkt_len > MAX_JUMBO_PKT_LEN)
		return -1;

	overhead_len = eth_dev_get_overhead_len(dev_info->max_rx_pktlen,
			dev_info->max_mtu);
	conf->rxmode.mtu = max_pkt_len - overhead_len;

	if (conf->rxmode.mtu > RTE_ETHER_MTU)
		conf->txmode.offloads |= RTE_ETH_TX_OFFLOAD_MULTI_SEGS;

	return 0;
}
#endif

#if 0
/*
 * Initializes a given port using global settings and with the RX buffers
 * coming from the mbuf_pool passed as a parameter.
 */
static inline int
port_init(uint16_t port, struct rte_mempool *mbuf_pool)
{
	struct rte_eth_conf port_conf = port_conf_default;
	const uint16_t rx_rings = 1, tx_rings = 1;
	uint16_t nb_rxd = RX_RING_SIZE;
	uint16_t nb_txd = TX_RING_SIZE;
	int retval;
	uint16_t q;
	struct rte_eth_dev_info dev_info;


	if (!rte_eth_dev_is_valid_port(port))
		return -1;

	nb_rx_queue = get_port_n_rx_queues(portid);
	n_tx_queue = nb_lcores;
	if (n_tx_queue > MAX_TX_QUEUE_PER_PORT)
		n_tx_queue = MAX_TX_QUEUE_PER_PORT;
	printf("Creating queues: nb_rxq=%d nb_txq=%u... ",
		nb_rx_queue, (unsigned)n_tx_queue );

	rte_eth_dev_info_get(port, &dev_info);
	if (dev_info.tx_offload_capa & RTE_ETH_TX_OFFLOAD_MBUF_FAST_FREE)
		port_conf.txmode.offloads |=
			RTE_ETH_TX_OFFLOAD_MBUF_FAST_FREE;
#if RTE_VERSION >= RTE_VERSION_NUM(22,7,0,0)	
	retval = config_port_max_pkt_len(&port_conf, &dev_info);
	if (retval != 0)
		rte_exit(EXIT_FAILURE,
				"Invalid max packet length: %u (port %u)\n",
				max_pkt_len, port);
#endif
	port_conf.rx_adv_conf.rss_conf.rss_hf &=
			dev_info.flow_type_rss_offloads;

	if (dev_info.max_rx_queues == 1)
			port_conf.rxmode.mq_mode = RTE_ETH_MQ_RX_NONE;

	if (port_conf.rx_adv_conf.rss_conf.rss_hf !=
				port_conf_default.rx_adv_conf.rss_conf.rss_hf) {
		printf("Port %u modified RSS hash function based on hardware support,"
				"requested:%#"PRIx64" configured:%#"PRIx64"\n",
				port,
				port_conf_default.rx_adv_conf.rss_conf.rss_hf,
				port_conf.rx_adv_conf.rss_conf.rss_hf);
		}

	/* Configure the Ethernet device. */
	retval = rte_eth_dev_configure(port, rx_rings, tx_rings, &port_conf);
	if (retval != 0)
		return retval;

	retval = rte_eth_dev_adjust_nb_rx_tx_desc(port, &nb_rxd, &nb_txd);
	if (retval != 0)
		return retval;

	/* init one TX queue per couple (lcore,port) */
	queueid = 0;
	for (lcore_id = 0; lcore_id < RTE_MAX_LCORE; lcore_id++) {
		if (rte_lcore_is_enabled(lcore_id) == 0)
			continue;

		if (numa_on)
			socketid =
			(uint8_t)rte_lcore_to_socket_id(lcore_id);
		else
			socketid = 0;

		printf("txq=%u,%d,%d ", lcore_id, queueid, socketid);
		fflush(stdout);

		txconf = &dev_info.default_txconf;
		txconf->offloads = local_port_conf.txmode.offloads;
		ret = rte_eth_tx_queue_setup(portid, queueid, nb_txd,
					     socketid, txconf);
		if (ret < 0)
			rte_exit(EXIT_FAILURE,
				"rte_eth_tx_queue_setup: err=%d, "
				"port=%d\n", ret, portid);

		qconf = &lcore_conf[lcore_id];
		qconf->tx_queue_id[portid] = queueid;
		queueid++;

		qconf->tx_port_id[qconf->n_tx_port] = portid;
		qconf->n_tx_port++;
	}
	printf("\n");

	/* Start the Ethernet port. */
	retval = rte_eth_dev_start(port);
	if (retval < 0)
		return retval;

	/* Display the port MAC address. */
	struct rte_ether_addr addr;
	rte_eth_macaddr_get(port, &addr);
	printf("Port %u MAC: %02" PRIx8 " %02" PRIx8 " %02" PRIx8
			   " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 "\n",
			port,
			addr.addr_bytes[0], addr.addr_bytes[1],
			addr.addr_bytes[2], addr.addr_bytes[3],
			addr.addr_bytes[4], addr.addr_bytes[5]);

	/* Enable RX in promiscuous mode for the Ethernet device. */
	retval = rte_eth_promiscuous_enable(port);
	if (retval < 0)
		return retval;

	return 0;
}
#endif




void print_bpm_packet(struct rt_fpga_packet *mpacket) 
{
#if 1
	printf("Dest 0x%02x:0x%02x:0x%02x:0x%02x:0x%02x:0x%02x\n", 
		mpacket->eth_cpu_mac[0],mpacket->eth_cpu_mac[1],mpacket->eth_cpu_mac[2], 
		mpacket->eth_cpu_mac[3],mpacket->eth_cpu_mac[4],mpacket->eth_cpu_mac[5]);
	printf("FPGA 0x%02x:0x%02x:0x%02x:0x%02x:0x%02x:0x%02x\n", 
		mpacket->eth_fpga_mac[0],mpacket->eth_fpga_mac[1],mpacket->eth_fpga_mac[2],
		mpacket->eth_fpga_mac[3],mpacket->eth_fpga_mac[4],mpacket->eth_fpga_mac[5]);
	printf("CPU IP: %d.%d.%d.%d\n",mpacket->ip_cpu_ip[0],mpacket->ip_cpu_ip[1],
		mpacket->ip_cpu_ip[2],mpacket->ip_cpu_ip[3]);
	printf("CPU UDP PORT: %d\n",mpacket->udp_cpu_port);
	printf("FPGA IP: %d.%d.%d.%d\n",mpacket->ip_fpga_ip[0],mpacket->ip_fpga_ip[1],
		mpacket->ip_fpga_ip[2],mpacket->ip_fpga_ip[3]);
	printf("FPGA UDP PORT: %d\n",mpacket->udp_fpga_port);
			
	uint8_t *e = (uint8_t *) mpacket;
	printf("Hex:\n%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8"\n"
		"%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8"\n"
		"%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8"\n"
		"%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8"\n"
		"%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8"\n"
		"%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8"\n"
		"%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8"\n"
		"%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8"\n"
		"%02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 "\n",		e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7], e[8], e[9],
		e[10], e[11], e[12], e[13], e[14], e[15], e[16], e[17], e[18], e[19],
		e[20], e[21], e[22], e[23], e[24], e[25], e[26], e[27], e[28], e[29],
		e[30], e[31], e[32], e[33], e[34], e[35], e[36], e[37], e[38], e[39],
		e[40], e[41], e[42], e[43], e[44], e[45], e[46], e[47], e[48], e[49],
		e[50], e[51], e[52], e[53], e[54], e[55], e[56], e[57], e[58], e[59],
		e[60], e[61], e[62], e[63], e[64], e[65], e[66], e[67], e[68], e[69],
		e[70], e[71], e[72], e[73], e[74], e[75], e[76], e[77], e[78], e[79],
		e[80], e[81], e[82], e[83]
		);
#endif
}



#if 0
/*
 * The lcore main. This is the main thread that does the work, reading from
 * an input port and writing to an output port.
 */
static __attribute__((noreturn)) void
lcore_main(void)
#endif
/* main processing loop */
int
main_loop(__attribute__((unused)) void *dummy)
{
	uint16_t port;
	struct rte_ether_addr addr;
	struct rt_fpga_packet *mpacket;
	unsigned lcore_id;
	int i;
	uint8_t queueid;
	uint16_t portid;
	uint16_t stat_id;
	struct lcore_conf *qconf;
	long long int cnt = 0;
	long long int spur_found = 0;
	long long int start_monitor_time = 0, end_monitor_time = 0;
	long long int stat_packet[9] = {0,0,0,0,0,0,0,0,0};

	lcore_id = rte_lcore_id();
	qconf = &lcore_conf[lcore_id];
	stat_id = qconf->stat_id;

	if (qconf->n_rx_queue == 0) {
		printf("lcore %u has nothing to do\n", lcore_id);
		return 0;
	}

	printf("entering main loop on lcore %u\n", lcore_id);

	for (i = 0; i < qconf->n_rx_queue; i++) {

		portid = qconf->rx_queue_list[i].port_id;
		queueid = qconf->rx_queue_list[i].queue_id;
		printf(" -- lcoreid=%u portid=%u rxqueueid=%hhu\n",
			lcore_id, portid, queueid);
		/*
		 * Check that the port is on the same NUMA node as the polling thread
		 * for best performance.
		 */

		if (rte_eth_dev_socket_id(portid) > 0 &&
				rte_eth_dev_socket_id(portid) !=
						(int)rte_socket_id())
			rte_exit(EXIT_FAILURE, "ERROR, port %u is on remote NUMA node to "
					"polling thread.\n\tPerformance will "
					"not be optimal.\n", portid);
	}


	memset(bpm_packet,0,sizeof(struct rt_fpga_packet) * 3);

	start_monitor_time = (rte_rdtsc_precise() / cxus);

	/* Run until the application is quit or killed. */
	while (!force_quit) {
		long long int start_loop = 0, end_loop = 0, start_packet_time = 0, end_packet_time = 0, start_free_time = 0, end_free_time = 0;

		if(likely(cnt > warmup_cycles))
			start_loop = (rte_rdtsc_precise() / cxus);

		for (i = 0; i < qconf->n_rx_queue; ++i) {
			portid = qconf->rx_queue_list[i].port_id;
			queueid = qconf->rx_queue_list[i].queue_id;
			rte_eth_macaddr_get(portid, &addr);

			/* Get burst of RX packets, from first port of pair. */
			struct rte_mbuf *bufs[32/*max 32 rx + 32tx*/] __rte_cache_aligned;

			/*
			static uint16_t nb_rx = rte_eth_rx_burst (uint16_t port_id, uint16_t queue_id, struct rte_mbuf **rx_pkts, const uint16_t nb_pkts) 
			 port_id: 	The port identifier of the Ethernet device.
			 queue_id:	The index of the receive queue from which to retrieve input packets. 
					The value must be in the range [0, nb_rx_queue - 1] previously supplied to rte_eth_dev_configure().
			 rx_pkts:	The address of an array of pointers to rte_mbuf structures that must be large enough to store nb_pkts pointers in it.
			 nb_pkts:	The maximum number of packets to retrieve. The value must be divisible by 8 in order to work with any driver. 
					Some drivers using vector instructions require that nb_pkts is divisible by 4 or 8, depending on the driver implementation.

			 nb_rx:		Number of returned packets
			*/

			long long int start_rx_burst = 0, end_rx_burst = 0;
 
			if(likely(cnt > warmup_cycles)) {
				start_rx_burst = (rte_rdtsc() / cxus);
			}

			const uint16_t nb_rx = rte_eth_rx_burst(portid, queueid,
					bufs, 8/*TODO: NOT RECEIVING ANYTHING WITH LESS THEN 8*/);

#if 0
			if(nb_rx > 0)
			{
				printf("RECEIVED %d on lcoreid=%u portid=%u rxqueueid=%hhu\n", nb_rx,
					lcore_id, portid, queueid);	
				print_bpm_packet(&bpm_packet[0]);
				printf("--------------------------------------------------\n");
			}
#endif
#if 0
			if(unlikely(nb_rx > 1))
			{
				printf("RECEIVED %d on lcoreid=%u portid=%u rxqueueid=%hhu\n", nb_rx,
					lcore_id, portid, queueid);	
				print_bpm_packet(&bpm_packet[0]);
				printf("--------------------------------------------------\n");
			}
#endif

			if(likely(cnt > warmup_cycles)) {
				end_rx_burst = (rte_rdtsc() / cxus);
			}

			if(likely(cnt > warmup_cycles)) {		
				if ((end_rx_burst - start_rx_burst) > qstat[stat_id].rte_eth_rx_burst_max_time) {			
					qstat[stat_id].rte_eth_rx_burst_max_time = (end_rx_burst - start_rx_burst);
				}
			}

			if(likely(cnt > warmup_cycles)) {
				start_packet_time = (rte_rdtsc() / cxus);
			}

			for (size_t i = 0; i < nb_rx; i++) {

				memcpy(&bpm_packet[0],&bpm_packet[1],sizeof(struct rt_fpga_packet));
				memcpy(&bpm_packet[1],&bpm_packet[2],sizeof(struct rt_fpga_packet));
				mpacket = rte_pktmbuf_mtod(bufs[i], struct rt_fpga_packet *);	
				memcpy(&bpm_packet[2],mpacket,sizeof(struct rt_fpga_packet));

				old_seq = cur_seq;
				cur_seq = mpacket->seq;

				// mark time on the first packet
				if(likely(cnt > warmup_cycles)) {

					qstat[stat_id].old_timestamp_tsc = qstat[stat_id].new_timestamp_tsc;


					if (i == 0) {
						qstat[stat_id].new_timestamp_tsc = (rte_rdtsc() / cxus);
					}

					if (qstat[stat_id].old_timestamp_tsc != 0)  {
						qstat[stat_id].tmp_time = qstat[stat_id].new_timestamp_tsc - qstat[stat_id].old_timestamp_tsc;
					

						if (cur_seq != ((old_seq + 1) & 0xffff)) {
							qstat[stat_id].lost_packet += 1;
//#define _PRINT_ERR_SEQ
#ifdef _PRINT_ERR_SEQ
							print_bpm_packet(&bpm_packet[0]);
							print_bpm_packet(&bpm_packet[1]);
							print_bpm_packet(&bpm_packet[2]);
#endif						

						}

						if (qstat[stat_id].tmp_time < qstat[stat_id].min_time)
							qstat[stat_id].min_time = qstat[stat_id].tmp_time;
						if (qstat[stat_id].tmp_time > qstat[stat_id].max_time)
							qstat[stat_id].max_time = qstat[stat_id].tmp_time;
						if (qstat[stat_id].max_num_packet < nb_rx)
							qstat[stat_id].max_num_packet = nb_rx; 
					
					}
					
					qstat[stat_id].rx_packets++;


					if (spur_found == 1) {
						spur_found = 0;
#ifdef _PRINT_ERR_SEQ
						print_bpm_packet(&bpm_packet[0]);
						print_bpm_packet(&bpm_packet[1]);
						print_bpm_packet(&bpm_packet[2]);
#endif
					}

				}



			}

		

			if(likely(cnt > warmup_cycles)) {
				end_packet_time = (rte_rdtsc() / cxus);
				if ((end_packet_time - start_packet_time) > qstat[stat_id].max_packet_time) {			
					qstat[stat_id].max_packet_time = (end_packet_time - start_packet_time);
				}
			}
			if(likely(cnt > warmup_cycles)) {
				start_free_time = (rte_rdtsc() / cxus);
			}

			for(size_t ind=0; ind<nb_rx; ind++)
			{
				rte_pktmbuf_free(bufs[ind]);
			}

			if(likely(cnt > warmup_cycles)) {
				end_free_time = (rte_rdtsc() / cxus);
				if ((end_free_time - start_free_time) > qstat[stat_id].max_free_time) {		
					qstat[stat_id].max_free_time = (end_free_time - start_free_time);
				}
			}

			if(likely(cnt > warmup_cycles)) {
				if (nb_rx > 0) {
					if (nb_rx <= 8)
						stat_packet[nb_rx-1] += 1;
					else
						stat_packet[nb_rx] += 1;
				}
			}				

			if (unlikely(nb_rx == 0))
				continue;
			else
				cnt++;

			


		}


		if ((cnt > warmup_cycles) && (start_loop != 0)) {
			end_loop = (rte_rdtsc_precise() / cxus);
			if ((end_loop - start_loop) > qstat[stat_id].max_loop_time) {			
				qstat[stat_id].max_loop_time = (end_loop - start_loop);
			}
		}

		end_monitor_time = (rte_rdtsc_precise() / cxus);

		if ((stop_flag) || ((end_monitor_time - start_monitor_time) > 10000000)) {
			stop_flag = false;
			cnt = 0;
		
			printf("lcore_id:%d  min_per:%lldus, max_per:%lldus, max_event_pkt=%lld, lost_pkt=%lld, spur_pkt=%lld, rx_pkt=%lld, "
				"rte_eth_rx_burst_max_time=%lldus max_loop_time=%lldus max_rx_packet_time=%lldus max_free_time=%lldus perc_one_pkt=%.3f%%\n",
				lcore_id,
				qstat[stat_id].min_time, qstat[stat_id].max_time,
				qstat[stat_id].max_num_packet,qstat[stat_id].lost_packet,
				qstat[stat_id].spurious_packet,qstat[stat_id].rx_packets,
				qstat[stat_id].rte_eth_rx_burst_max_time,qstat[stat_id].max_loop_time,
				qstat[stat_id].max_packet_time,qstat[stat_id].max_free_time,(double)stat_packet[0]*100/(double)qstat[stat_id].rx_packets);
			/*
			for (int i = 0; i < 8; i++) {
				printf("events %lld: %d\n",i+1,stat_packet[i]);
			}
			printf("events > 8: %lld\n",stat_packet[8]);
			*/
			qstat[stat_id].min_time = 1e10; qstat[stat_id].max_time = 0; qstat[stat_id].max_num_packet = 0, qstat[stat_id].lost_packet = 0;
			qstat[stat_id].spurious_packet = 0;qstat[stat_id].rx_packets = 0;qstat[stat_id].rte_eth_rx_burst_max_time = 0;
			qstat[stat_id].max_packet_time = 0;qstat[stat_id].max_free_time = 0;
			qstat[stat_id].max_loop_time = 0;
			start_loop = 0; end_loop = 0; 
			qstat[stat_id].old_timestamp_tsc = 0; qstat[stat_id].new_timestamp_tsc = 0;
			old_seq = 0; cur_seq = 0;
			start_monitor_time = (rte_rdtsc_precise() / cxus);

			for (int i = 0; i < 9; i++)
				stat_packet[i] = 0;

		}	
	

	}
	return 0;
}

/*
 * This expression is used to calculate the number of mbufs needed
 * depending on user input, taking  into account memory for rx and
 * tx hardware rings, cache per lcore and mtable per port per lcore.
 * RTE_MAX is used to ensure that NB_MBUF never goes below a minimum
 * value of 8192
 */
#define NB_MBUF(nports) RTE_MAX(	\
	(nports*nb_rx_queue*nb_rxd +		\
	nports*nb_lcores*MAX_PKT_BURST +	\
	nports*n_tx_queue*nb_txd +		\
	nb_lcores*MEMPOOL_CACHE_SIZE),		\
	(unsigned)8192)

/* Parse the argument given in the command line of the application */
static int
parse_args(int argc, char **argv)
{
	int opt, ret;
	char **argvopt;
	int option_index;
	char *prgname = argv[0];
	uint8_t lcore_params = 0;
	uint8_t eventq_sched = 0;
	uint8_t eth_rx_q = 0;

	argvopt = argv;

	/* Error or normal output strings. */
	while ((opt = getopt_long(argc, argvopt, short_options,
				lgopts, &option_index)) != EOF) {

		switch (opt) {
		/* portmask */
		case 'p':
			enabled_port_mask = parse_portmask(optarg);
			if (enabled_port_mask == 0) {
				fprintf(stderr, "Invalid portmask\n");
				print_usage(prgname);
				return -1;
			}
			break;

		/* long options */
		case CMD_LINE_OPT_CONFIG_NUM:
			ret = parse_config(optarg);
			if (ret) {
				fprintf(stderr, "Invalid config\n");
				print_usage(prgname);
				return -1;
			}
			lcore_params = 1;
			break;

		default:
			print_usage(prgname);
			return -1;
		}
	}

	if (optind >= 0)
		argv[optind-1] = prgname;

	ret = optind-1;
	optind = 1; /* reset getopt lib */
	return ret;
}

static int
init_mem(uint16_t portid, unsigned int nb_mbuf)
{
	struct lcore_conf *qconf;
	int socketid;
	unsigned lcore_id;
	char s[64];

	for (lcore_id = 0; lcore_id < RTE_MAX_LCORE; lcore_id++) {
		if (rte_lcore_is_enabled(lcore_id) == 0)
			continue;

		if (numa_on)
			socketid = rte_lcore_to_socket_id(lcore_id);
		else
			socketid = 0;

		if (socketid >= NB_SOCKETS) {
			rte_exit(EXIT_FAILURE,
				"Socket %d of lcore %u is out of range %d\n",
				socketid, lcore_id, NB_SOCKETS);
		}

		if (pktmbuf_pool[portid][socketid] == NULL) {
			snprintf(s, sizeof(s), "mbuf_pool_%d:%d",
				 portid, socketid);
			pktmbuf_pool[portid][socketid] =
				rte_pktmbuf_pool_create(s, nb_mbuf,
					MEMPOOL_CACHE_SIZE, 0,
					RTE_MBUF_DEFAULT_BUF_SIZE, socketid);
			if (pktmbuf_pool[portid][socketid] == NULL)
				rte_exit(EXIT_FAILURE,
					"Cannot init mbuf pool on socket %d\n",
					socketid);
			else
				printf("Allocated mbuf pool on socket %d\n",
					socketid);
#if 0
			/* Setup either LPM or EM(f.e Hash). But, only once per
			 * available socket.
			 */
			if (!lkp_per_socket[socketid]) {
				l3fwd_lkp.setup(socketid);
				lkp_per_socket[socketid] = 1;
			}
#endif
		}
		qconf = &lcore_conf[lcore_id];
#if 0	
		qconf->ipv4_lookup_struct =
			l3fwd_lkp.get_ipv4_lookup_struct(socketid);
		qconf->ipv6_lookup_struct =
			l3fwd_lkp.get_ipv6_lookup_struct(socketid);
#endif
	}
	return 0;
}

/* Check the link status of all ports in up to 9s, and print them finally */
static void
check_all_ports_link_status(uint32_t port_mask)
{
#define CHECK_INTERVAL 100 /* 100ms */
#define MAX_CHECK_TIME 300 /* 9s (90 * 100ms) in total */
	uint16_t portid;
	uint8_t count, all_ports_up, print_flag = 0;
	struct rte_eth_link link;
	int ret;

	printf("\nChecking link status");
	fflush(stdout);
	for (count = 0; count <= MAX_CHECK_TIME; count++) {
		if (force_quit)
			return;
		all_ports_up = 1;
		RTE_ETH_FOREACH_DEV(portid) {
			if (force_quit)
				return;
			if ((port_mask & (1 << portid)) == 0)
				continue;
			memset(&link, 0, sizeof(link));
			ret = rte_eth_link_get_nowait(portid, &link);
			if (ret < 0) {
				all_ports_up = 0;
				if (print_flag == 1)
					printf("Port %u link get failed: %s\n",
						portid, rte_strerror(-ret));
				continue;
			}
			/* print link status if flag set */
			if (print_flag == 1) {
				if (link.link_status)
					printf(
					"Port%d Link Up. Speed %u Mbps -%s\n",
						portid, link.link_speed,
				(link.link_duplex == ETH_LINK_FULL_DUPLEX) ?
					("full-duplex") : ("half-duplex"));
				else
					printf("Port %d Link Down\n", portid);
				continue;
			}
			/* clear all_ports_up flag if any link down */
			if (link.link_status == ETH_LINK_DOWN) {
				all_ports_up = 0;
				break;
			}
		}
		/* after finally printing all link status, get out */
		if (print_flag == 1)
			break;

		if (all_ports_up == 0) {
			printf(".");
			fflush(stdout);
			rte_delay_ms(CHECK_INTERVAL);
		}

		/* set the print_flag if all ports up or timeout */
		if (all_ports_up == 1 || count == (MAX_CHECK_TIME - 1)) {
			print_flag = 1;
			printf("done\n");
		}
	}
}

/* Init all statistics */
static void init_stat()
{
	unsigned i;

	for (i = 0; i < MAX_STAT; i++) {
		qstat[i].min_time = 1e10;
		qstat[i].max_time = 0;
		qstat[i].tmp_time = 0;
		qstat[i].max_num_packet = 0;
		qstat[i].lost_packet = 0;
		qstat[i].spurious_packet = 0;
		qstat[i].rx_packets = 0;
		qstat[i].old_timestamp_tsc = 0;
		qstat[i].new_timestamp_tsc = 0; 
		qstat[i].min_loop_time = 0;
		qstat[i].max_loop_time = 0;
		qstat[i].rte_eth_rx_burst_max_time = 0;
		qstat[i].max_packet_time = 0;
		qstat[i].max_free_time = 0;
	}
}


/*
 * The main function, which does initialization and calls the per-lcore
 * functions.
 */
int
main(int argc, char *argv[])
{
	struct rte_mempool *mbuf_pool;
	struct lcore_conf *qconf;
	struct rte_eth_dev_info dev_info;
	struct rte_eth_txconf *txconf;
	int ret;
	unsigned nb_ports;
	uint16_t queueid, portid;
	unsigned lcore_id;
	uint32_t n_tx_queue, nb_lcores;
	uint8_t nb_rx_queue, queue, socketid;

	force_quit = false;
	stop_flag = false;
	/* Associate signal_hanlder function with USR signals */
	signal(SIGUSR1, signal_handler);
	signal(SIGINT, signal_handler);
	signal(SIGTERM, signal_handler);

	init_stat();

	/* Initialize the Environment Abstraction Layer (EAL). */
	ret = rte_eal_init(argc, argv);
	if (ret < 0)
		rte_exit(EXIT_FAILURE, "Error with EAL initialization\n");
	argc -= ret;
	argv += ret;

	/* parse application arguments (after the EAL ones) */
	ret = parse_args(argc, argv);
	if (ret < 0)
		rte_exit(EXIT_FAILURE, "Invalid command line parameters\n");

	if (check_lcore_params() < 0)
		rte_exit(EXIT_FAILURE, "check_lcore_params failed\n");

	ret = init_lcore_rx_queues();
	if (ret < 0)
		rte_exit(EXIT_FAILURE, "init_lcore_rx_queues failed\n");

	nb_ports = rte_eth_dev_count_avail();

	if (check_port_config() < 0)
		rte_exit(EXIT_FAILURE, "check_port_config failed\n");

	nb_lcores = rte_lcore_count();

	printf("rte_lcore_count()=%d rte_eth_dev_count_avail())=%d\n", rte_lcore_count(), rte_eth_dev_count_avail());
	//if (rte_lcore_count() != 1 || rte_eth_dev_count_avail() != 1)
	//	printf("\nWARNING: Too many lcores/ports enabled. Only 1 used for each.\n");
#if 0
	/* Creates a new mempool in memory to hold the mbufs. */
	mbuf_pool = rte_pktmbuf_pool_create("MBUF_POOL", NUM_MBUFS * 1 /* only one port used */,
		MBUF_CACHE_SIZE, 0, RTE_MBUF_DEFAULT_BUF_SIZE, rte_socket_id());

	if (mbuf_pool == NULL)
		rte_exit(EXIT_FAILURE, "Cannot create mbuf pool\n");
#endif
	/* Initialize port */
	RTE_ETH_FOREACH_DEV(portid) {
		struct rte_eth_conf local_port_conf = port_conf_default;

		/* skip ports that are not enabled */
		if ((enabled_port_mask & (1 << portid)) == 0) {
			printf("\nSkipping disabled port %d\n", portid);
			continue;
		}

		/* init port */
		printf("Initializing port %d ... ", portid );
		fflush(stdout);

	//struct rte_eth_conf port_conf = port_conf_default;
	//const uint16_t rx_rings = 1, tx_rings = 1;
	//uint16_t nb_rxd = RX_RING_SIZE;
	//uint16_t nb_txd = TX_RING_SIZE;
	//int retval;
	//uint16_t q;
	//



		nb_rx_queue = get_port_n_rx_queues(portid);
		n_tx_queue = nb_lcores;
		if (n_tx_queue > MAX_TX_QUEUE_PER_PORT)
			n_tx_queue = MAX_TX_QUEUE_PER_PORT;
		printf("Creating queues: nb_rxq=%d nb_txq=%u... ",
			nb_rx_queue, (unsigned)n_tx_queue );

		ret = rte_eth_dev_info_get(portid, &dev_info);
		if (ret != 0)
			rte_exit(EXIT_FAILURE,
				"Error during getting device (port %u) info: %s\n",
				portid, strerror(-ret));
#if RTE_VERSION < RTE_VERSION_NUM(22,7,0,0)
		if (dev_info.tx_offload_capa & DEV_TX_OFFLOAD_MBUF_FAST_FREE)
			local_port_conf.txmode.offloads |=
				DEV_TX_OFFLOAD_MBUF_FAST_FREE;
#else			
		if (dev_info.tx_offload_capa & RTE_ETH_TX_OFFLOAD_MBUF_FAST_FREE)
			local_port_conf.txmode.offloads |=
				RTE_ETH_TX_OFFLOAD_MBUF_FAST_FREE;
#endif
				
#if RTE_VERSION >= RTE_VERSION_NUM(22,7,0,0)
		retval = config_port_max_pkt_len(&local_port_conf, &dev_info);
		if (retval != 0)
			rte_exit(EXIT_FAILURE,
					"Invalid max packet length: %u (port %u)\n",
					max_pkt_len, port);
#endif
		local_port_conf.rx_adv_conf.rss_conf.rss_hf &=
				dev_info.flow_type_rss_offloads;
#if RTE_VERSION >= RTE_VERSION_NUM(22,7,0,0)	
		if (dev_info.max_rx_queues == 1)
				local_port_conf.rxmode.mq_mode = RTE_ETH_MQ_RX_NONE;
#endif

		if (local_port_conf.rx_adv_conf.rss_conf.rss_hf !=
				port_conf_default.rx_adv_conf.rss_conf.rss_hf) {
			printf("Port %u modified RSS hash function based on hardware support,"
				"requested:%#"PRIx64" configured:%#"PRIx64"\n",
				portid,
				port_conf_default.rx_adv_conf.rss_conf.rss_hf,
				local_port_conf.rx_adv_conf.rss_conf.rss_hf);
		}

		/* Configure the Ethernet device. */
		ret = rte_eth_dev_configure(portid, nb_rx_queue,
					(uint16_t)n_tx_queue, &local_port_conf);
		if (ret < 0)
			rte_exit(EXIT_FAILURE,
				"Cannot configure device: err=%d, port=%d\n",
				ret, portid);

		ret = rte_eth_dev_adjust_nb_rx_tx_desc(portid, &nb_rxd,
						       &nb_txd);
		if (ret < 0)
			rte_exit(EXIT_FAILURE,
				 "Cannot adjust number of descriptors: err=%d, "
				 "port=%d\n", ret, portid);

		/* init memory */
		if (true/*!per_port_pool*/) {
			/* portid = 0; this is *not* signifying the first port,
			 * rather, it signifies that portid is ignored.
			 */
			ret = init_mem(0, NB_MBUF(nb_ports));
		} else {
			ret = init_mem(portid, NB_MBUF(1));
		}
		if (ret < 0)
			rte_exit(EXIT_FAILURE, "init_mem failed\n");

		/* init one TX queue per couple (lcore,port) */
		queueid = 0;
		for (lcore_id = 0; lcore_id < RTE_MAX_LCORE; lcore_id++) {
			if (rte_lcore_is_enabled(lcore_id) == 0)
				continue;

			if (numa_on)
				socketid =
				(uint8_t)rte_lcore_to_socket_id(lcore_id);
			else
				socketid = 0;

			printf("txq=%u,%d,%d struct lcore_conf *qconf;", lcore_id, queueid, socketid);
			fflush(stdout);

			txconf = &dev_info.default_txconf;
			txconf->offloads = local_port_conf.txmode.offloads;
			ret = rte_eth_tx_queue_setup(portid, queueid, nb_txd,
						     socketid, txconf);
			if (ret < 0)
				rte_exit(EXIT_FAILURE,
					"rte_eth_tx_queue_setup: err=%d, "
					"port=%d\n", ret, portid);

			qconf = &lcore_conf[lcore_id];
			qconf->tx_queue_id[portid] = queueid;
			queueid++;

			qconf->tx_port_id[qconf->n_tx_port] = portid;
			qconf->n_tx_port++;
		}
		printf("\n");
	}

	for (lcore_id = 0; lcore_id < RTE_MAX_LCORE; lcore_id++) {
		if (rte_lcore_is_enabled(lcore_id) == 0)
			continue;
		qconf = &lcore_conf[lcore_id];
		printf("\nInitializing rx queues on lcore %u ... ", lcore_id );
		fflush(stdout);
		/* init RX queues */
		for(queue = 0; queue < qconf->n_rx_queue; ++queue) {
			printf("looping queue %d/%d\n",queue, qconf->n_rx_queue);
			struct rte_eth_rxconf rxq_conf;

			portid = qconf->rx_queue_list[queue].port_id;
			queueid = qconf->rx_queue_list[queue].queue_id;

			if (numa_on)
				socketid =
				(uint8_t)rte_lcore_to_socket_id(lcore_id);
			else
				socketid = 0;

			printf("rxq=%d,%d,%d ", portid, queueid, socketid);
			fflush(stdout);

			ret = rte_eth_dev_info_get(portid, &dev_info);
			if (ret != 0)
				rte_exit(EXIT_FAILURE,
					"Error during getting device (port %u) info: %s\n",
					portid, strerror(-ret));

			rxq_conf = dev_info.default_rxconf;
			rxq_conf.offloads = port_conf_default.rxmode.offloads;
			if (true/*!per_port_pool*/) //TODO: separate buffer pools per port; disabled by default
				ret = rte_eth_rx_queue_setup(portid, queueid,
						nb_rxd, socketid,
						&rxq_conf,
						pktmbuf_pool[0][socketid]);
			else
				ret = rte_eth_rx_queue_setup(portid, queueid,
						nb_rxd, socketid,
						&rxq_conf,
						pktmbuf_pool[portid][socketid]);
			if (ret < 0)
				rte_exit(EXIT_FAILURE,
				"rte_eth_rx_queue_setup: err=%d, port=%d\n",
				ret, portid);
		}
	}

	printf("\n");

	/* start ports */
	RTE_ETH_FOREACH_DEV(portid) {
		if ((enabled_port_mask & (1 << portid)) == 0) {
			continue;
		}
		/* Start device */
		ret = rte_eth_dev_start(portid);
		if (ret < 0)
			rte_exit(EXIT_FAILURE,
				"rte_eth_dev_start: err=%d, port=%d\n",
				ret, portid);

		/* Display the port MAC address. */
		struct rte_ether_addr addr;
		rte_eth_macaddr_get(portid, &addr);
		printf("Port %u MAC: %02" PRIx8 " %02" PRIx8 " %02" PRIx8
				   " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 "\n",
				portid,
				addr.addr_bytes[0], addr.addr_bytes[1],
				addr.addr_bytes[2], addr.addr_bytes[3],
				addr.addr_bytes[4], addr.addr_bytes[5]);
		/*
		 * If enabled, put device in promiscuous mode.
		 * This allows IO forwarding mode to forward packets
		 * to itself through 2 cross-connected  ports of the
		 * target machine.
		 */
		/* Enable RX in promiscuous mode for the Ethernet device. */
		if (true/*promiscuous_on*/) {
			ret = rte_eth_promiscuous_enable(portid);
			if (ret != 0)
				rte_exit(EXIT_FAILURE,
					"rte_eth_promiscuous_enable: err=%s, port=%u\n",
					rte_strerror(-ret), portid);
		}
	}

	printf("\n");

	for (lcore_id = 0; lcore_id < RTE_MAX_LCORE; lcore_id++) {
		if (rte_lcore_is_enabled(lcore_id) == 0)
			continue;
		qconf = &lcore_conf[lcore_id];
		for (queue = 0; queue < qconf->n_rx_queue; ++queue) {
			portid = qconf->rx_queue_list[queue].port_id;
			queueid = qconf->rx_queue_list[queue].queue_id;
#if 0
			if (prepare_ptype_parser(portid, queueid) == 0)
				rte_exit(EXIT_FAILURE, "ptype check fails\n");
#endif
		}
	}


	//check_all_ports_link_status(enabled_port_mask);

	ret = 0;
	/* launch per-lcore init on every lcore */
	rte_eal_mp_remote_launch(main_loop, NULL, CALL_MASTER);
	RTE_LCORE_FOREACH_SLAVE(lcore_id) {
		if (rte_eal_wait_lcore(lcore_id) < 0) {
			ret = -1;
			break;
		}
	}

	/* stop ports */
	RTE_ETH_FOREACH_DEV(portid) {
		if ((enabled_port_mask & (1 << portid)) == 0)
			continue;
		printf("Closing port %d...", portid);
		rte_eth_dev_stop(portid);
		rte_eth_dev_close(portid);
		printf(" Done\n");
	}

	/* clean up the EAL */
	rte_eal_cleanup();

	printf("Bye...\n");

#if 0
	/* Call lcore_main on the master core only. */
	lcore_main();
#endif
	return ret;
}
